/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __GLQUAKE_H
#define __GLQUAKE_H

// for restoring post-transform
extern glmatrix r_world_matrix;

// partitioned depth range slices for 2d/3d stuff
#define QGL_DEPTH_2D_BEGIN		0.0f
#define QGL_DEPTH_2D_END		0.005f
#define QGL_DEPTH_3D_BEGIN		QGL_DEPTH_2D_END
#define QGL_DEPTH_3D_END		1.0f
#define QGL_DEPTH_VM_END		0.3f

void GL_BeginRendering (int *width, int *height);
void GL_EndRendering (void);
void GL_Set2D (void);

// johnfitz -- removed texture object stuff since they are standard in gl 1.1

extern	int glwidth, glheight;

// r_local.h -- private refresh defs

#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
// normalizing factor so player model works out to about
//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		480

#define TILE_SIZE		128		// size of textures generated by R_GenTiledSurf

#define SKYSHIFT		7
#define	SKYSIZE			(1 << SKYSHIFT)
#define SKYMASK			(SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01


void R_TimeRefresh_f (void);
void R_ReadPointFile_f (void);
texture_t *R_TextureAnimation (texture_t *base, int frame);

typedef struct surfcache_s
{
	struct surfcache_s	*next;
	struct surfcache_s 	**owner;		// NULL is an empty chunk of memory
	int					lightadj[MAX_SURFACE_STYLES]; // checked for strobe flush
	int					dlight;
	int					size;		// including header
	unsigned			width;
	unsigned			height;		// DEBUG only needed for debug
	float				mipscale;
	struct texture_s	*texture;	// checked for animating textures
	byte				data[4];	// width*height elements
} surfcache_t;


typedef struct
{
	byte		*surfdat;	// destination for generated surface
	int			rowbytes;	// destination logical width in bytes
	msurface_t	*surf;		// description for surface to generate
	fixed8_t	lightadj[MAX_SURFACE_STYLES];
	// adjust for lightmap levels for dynamic lighting
	texture_t	*texture;	// corrected for animating textures
	int			surfmip;	// mipmapped ratio of surface texels / world pixels
	int			surfwidth;	// in mipmapped texels
	int			surfheight;	// in mipmapped texels
} drawsurf_t;


// this is needed outside of r_part now...
// these aren't really "types" but just relate to spawn positions and sorting
typedef struct particle_type_s
{
	struct particle_s *particles;
	int numparticles;
	vec3_t spawnorg;
	struct particle_type_s *next;
} particle_type_t;


//====================================================

extern	qboolean	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;

//
// view origin
//
extern	vec3_t	vup;
extern	vec3_t	vpn;
extern	vec3_t	vright;
extern	vec3_t	r_origin;

//
// screen size info
//
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	int		d_lightstylevalue[256];	// 8.8 fraction of base light value

extern	qboolean	envmap;

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_speeds;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_fullbright;
extern	cvar_t	r_lightmap;
extern	cvar_t	r_shadows;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;

extern	cvar_t	gl_clear;
extern	cvar_t	gl_cull;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_nocolors;

extern	cvar_t	gl_max_size;
extern	cvar_t	gl_playermip;

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;

// Multitexture
// johnfitz -- modified multitexture support
extern PFNGLMULTITEXCOORD2FARBPROC  qglMultiTexCoord2fARB;
extern PFNGLACTIVETEXTUREARBPROC    qglActiveTexture;
extern PFNGLCLIENTACTIVETEXTUREARBPROC    qglClientActiveTexture;
// johnfitz

// johnfitz -- anisotropic filtering
#define	GL_TEXTURE_MAX_ANISOTROPY_EXT 0x84FE
#define	GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT 0x84FF
// johnfitz

// johnfitz -- GL_EXT_texture_env_combine
//the values for GL_ARB_ are identical
// of course if you read the documentation you'll know that these aren't actually extensions any more...
#define GL_COMBINE			0x8570
#define GL_COMBINE_RGB		0x8571
#define GL_COMBINE_ALPHA	0x8572
#define GL_RGB_SCALE		0x8573
#define GL_CONSTANT			0x8576
#define GL_PRIMARY_COLOR	0x8577
#define GL_PREVIOUS			0x8578
#define GL_SOURCE0_RGB		0x8580
#define GL_SOURCE1_RGB		0x8581
#define GL_SOURCE0_ALPHA	0x8588
#define GL_SOURCE1_ALPHA	0x8589

void GL_BlendTest (void);


// johnfitz -- rendering statistics
extern int rs_brushpolys, rs_aliaspolys, rs_drawelements;
extern int rs_dynamiclightmaps;
// johnfitz

// johnfitz -- track developer statistics that vary every frame
extern cvar_t devstats;
typedef struct
{
	int		packetsize;
	int		edicts;
	int		visedicts;
	int		efrags;
	int		tempents;
	int		beams;
	int		dlights;
} devstats_t;
extern devstats_t dev_stats, dev_peakstats;
// johnfitz

//ohnfitz -- reduce overflow warning spam
typedef struct
{
	float	packetsize;
	float	efrags;
	float	beams;
} overflowtimes_t;
extern overflowtimes_t dev_overflows; //this stores the last time overflow messages were displayed, not the last time overflows occured
#define CONSOLE_RESPAM_TIME 3 // seconds between repeated warning messages
// johnfitz

// johnfitz -- moved here from r_brush.c
#define MAX_LIGHTMAPS 256 // johnfitz -- was 64
// johnfitz

extern qboolean r_drawflat_cheatsafe, r_fullbright_cheatsafe, r_lightmap_cheatsafe, r_drawworld_cheatsafe; // johnfitz

// johnfitz -- fog functions called from outside gl_fog.c
void Fog_ParseServerMessage (void);
float *Fog_GetColor (void);
float Fog_GetDensity (void);
void Fog_EnableGFog (void);
void Fog_DisableGFog (void);
void Fog_StartAdditive (void);
void Fog_StopAdditive (void);
void Fog_SetupFrame (void);
void Fog_NewMap (void);
void Fog_Init (void);
// johnfitz

void R_NewGame (void);

void R_AnimateLight (void);
void R_MarkLeafs (void);
qboolean R_CullBox (vec3_t emins, vec3_t emaxs);
qboolean R_CullSphere (const float *center, const float radius, const int clipflags);
void R_StoreEfrags (efrag_t **ppefrag);
qboolean R_CullModelForEntity (entity_t *e);
void R_MarkLights (dlight_t *light, int num, mnode_t *node);

void R_InitParticles (void);
void R_DrawParticles (void);
void CL_RunParticles (void);
void R_ClearParticles (void);

void R_TranslatePlayerSkin (int playernum);
void R_TranslateNewPlayerSkin (int playernum); // johnfitz -- this handles cases when the actual texture changes
void R_UpdateWarpTextures (void);

void R_DrawWorld (void);
void R_DrawBrushModel (entity_t *e);
void R_DrawSpriteModel (entity_t *e);
void R_DrawParticlesForType (particle_type_t *pt);

void GL_BuildLightmaps (void);

int R_LightPoint (entity_t *e);

void R_SubdivideSurface (model_t *mod, msurface_t *surf);
void R_BuildLightMap (msurface_t *surf, byte *dest, int stride);
void R_RenderDynamicLightmaps (msurface_t *surf);

void GL_DrawAliasShadow (entity_t *e);
void GL_MakeAliasModelDisplayLists (stvert_t *stverts, dtriangle_t *triangles);

void Sky_Init (void);
void Sky_NewMap (void);
void Sky_LoadTexture (texture_t *mt, byte *data);
void Sky_LoadSkyBox (char *name);

// mh - GL_RGB doesn't actually exist in hardware...
// http://www.opengl.org/wiki/Common_Mistakes#Texture_upload_and_pixel_reads
// And if you are interested, most GPUs like chunks of 4 bytes. In other words, RGBA or BGRA is prefered.
// RGB and BGR is considered bizarre since most GPUs, most CPUs and any other kind of chip don't handle 24
// bits. This means, the driver converts your RGB or BGR to what the GPU prefers, which typically is BGRA.
// also note: http://www.opengl.org/wiki/Common_Mistakes#Paletted_textures
// Notice that the format is GL_BGRA. As explain before, most GPUs prefer the BGRA format; using RGB,
// BGR and RGBA results in lower performance.
#ifndef GL_BGRA
#define GL_BGRA  0x80E1
#endif

void GL_PixelStore (int rowalign, int rowlen);

#define BYTE_CLAMP(a) ((a)>255?255:((a)<0?0:(a)))
#define BYTE_CLAMPF(a) ((a)>1?255:((a)<0?0:(a) * 255.0f))

//// ================================================================================================================================= ////
////	EXTENSIONS
//// ================================================================================================================================= ////

// occlusion queries
extern PFNGLGENQUERIESPROC qglGenQueries;
extern PFNGLDELETEQUERIESPROC qglDeleteQueries;
extern PFNGLISQUERYPROC qglIsQuery;
extern PFNGLBEGINQUERYPROC qglBeginQuery;
extern PFNGLENDQUERYPROC qglEndQuery;
extern PFNGLGETQUERYIVPROC qglGetQueryiv;
extern PFNGLGETQUERYOBJECTIVPROC qglGetQueryObjectiv;
extern PFNGLGETQUERYOBJECTUIVPROC qglGetQueryObjectuiv;

// vbo/pbo
extern PFNGLBINDBUFFERPROC qglBindBuffer;
extern PFNGLDELETEBUFFERSPROC qglDeleteBuffers;
extern PFNGLGENBUFFERSPROC qglGenBuffers;
extern PFNGLISBUFFERPROC qglIsBuffer;
extern PFNGLBUFFERDATAPROC qglBufferData;
extern PFNGLBUFFERSUBDATAPROC qglBufferSubData;
extern PFNGLGETBUFFERSUBDATAPROC qglGetBufferSubData;
extern PFNGLMAPBUFFERPROC qglMapBuffer;
extern PFNGLUNMAPBUFFERPROC qglUnmapBuffer;
extern PFNGLGETBUFFERPARAMETERIVPROC qglGetBufferParameteriv;
extern PFNGLGETBUFFERPOINTERVPROC qglGetBufferPointerv;

// gl 1.5 defines
#define GL_BUFFER_SIZE                    0x8764
#define GL_BUFFER_USAGE                   0x8765
#define GL_QUERY_COUNTER_BITS             0x8864
#define GL_CURRENT_QUERY                  0x8865
#define GL_QUERY_RESULT                   0x8866
#define GL_QUERY_RESULT_AVAILABLE         0x8867
#define GL_ARRAY_BUFFER                   0x8892
#define GL_ELEMENT_ARRAY_BUFFER           0x8893
#define GL_ARRAY_BUFFER_BINDING           0x8894
#define GL_ELEMENT_ARRAY_BUFFER_BINDING   0x8895
#define GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING 0x889F
#define GL_READ_ONLY                      0x88B8
#define GL_WRITE_ONLY                     0x88B9
#define GL_READ_WRITE                     0x88BA
#define GL_BUFFER_ACCESS                  0x88BB
#define GL_BUFFER_MAPPED                  0x88BC
#define GL_BUFFER_MAP_POINTER             0x88BD
#define GL_STREAM_DRAW                    0x88E0
#define GL_STREAM_READ                    0x88E1
#define GL_STREAM_COPY                    0x88E2
#define GL_STATIC_DRAW                    0x88E4
#define GL_STATIC_READ                    0x88E5
#define GL_STATIC_COPY                    0x88E6
#define GL_DYNAMIC_DRAW                   0x88E8
#define GL_DYNAMIC_READ                   0x88E9
#define GL_DYNAMIC_COPY                   0x88EA
#define GL_SAMPLES_PASSED                 0x8914

// gl 2.0 defines
#define GL_BLEND_EQUATION_RGB             GL_BLEND_EQUATION
#define GL_VERTEX_ATTRIB_ARRAY_ENABLED    0x8622
#define GL_VERTEX_ATTRIB_ARRAY_SIZE       0x8623
#define GL_VERTEX_ATTRIB_ARRAY_STRIDE     0x8624
#define GL_VERTEX_ATTRIB_ARRAY_TYPE       0x8625
#define GL_CURRENT_VERTEX_ATTRIB          0x8626
#define GL_VERTEX_PROGRAM_POINT_SIZE      0x8642
#define GL_VERTEX_PROGRAM_TWO_SIDE        0x8643
#define GL_VERTEX_ATTRIB_ARRAY_POINTER    0x8645
#define GL_STENCIL_BACK_FUNC              0x8800
#define GL_STENCIL_BACK_FAIL              0x8801
#define GL_STENCIL_BACK_PASS_DEPTH_FAIL   0x8802
#define GL_STENCIL_BACK_PASS_DEPTH_PASS   0x8803
#define GL_MAX_DRAW_BUFFERS               0x8824
#define GL_DRAW_BUFFER0                   0x8825
#define GL_DRAW_BUFFER1                   0x8826
#define GL_DRAW_BUFFER2                   0x8827
#define GL_DRAW_BUFFER3                   0x8828
#define GL_DRAW_BUFFER4                   0x8829
#define GL_DRAW_BUFFER5                   0x882A
#define GL_DRAW_BUFFER6                   0x882B
#define GL_DRAW_BUFFER7                   0x882C
#define GL_DRAW_BUFFER8                   0x882D
#define GL_DRAW_BUFFER9                   0x882E
#define GL_DRAW_BUFFER10                  0x882F
#define GL_DRAW_BUFFER11                  0x8830
#define GL_DRAW_BUFFER12                  0x8831
#define GL_DRAW_BUFFER13                  0x8832
#define GL_DRAW_BUFFER14                  0x8833
#define GL_DRAW_BUFFER15                  0x8834
#define GL_BLEND_EQUATION_ALPHA           0x883D
#define GL_POINT_SPRITE                   0x8861
#define GL_COORD_REPLACE                  0x8862
#define GL_MAX_VERTEX_ATTRIBS             0x8869
#define GL_VERTEX_ATTRIB_ARRAY_NORMALIZED 0x886A
#define GL_MAX_TEXTURE_COORDS             0x8871
#define GL_MAX_TEXTURE_IMAGE_UNITS        0x8872
#define GL_FRAGMENT_SHADER                0x8B30
#define GL_VERTEX_SHADER                  0x8B31
#define GL_MAX_FRAGMENT_UNIFORM_COMPONENTS 0x8B49
#define GL_MAX_VERTEX_UNIFORM_COMPONENTS  0x8B4A
#define GL_MAX_VARYING_FLOATS             0x8B4B
#define GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS 0x8B4C
#define GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS 0x8B4D
#define GL_SHADER_TYPE                    0x8B4F
#define GL_FLOAT_VEC2                     0x8B50
#define GL_FLOAT_VEC3                     0x8B51
#define GL_FLOAT_VEC4                     0x8B52
#define GL_INT_VEC2                       0x8B53
#define GL_INT_VEC3                       0x8B54
#define GL_INT_VEC4                       0x8B55
#define GL_BOOL                           0x8B56
#define GL_BOOL_VEC2                      0x8B57
#define GL_BOOL_VEC3                      0x8B58
#define GL_BOOL_VEC4                      0x8B59
#define GL_FLOAT_MAT2                     0x8B5A
#define GL_FLOAT_MAT3                     0x8B5B
#define GL_FLOAT_MAT4                     0x8B5C
#define GL_SAMPLER_1D                     0x8B5D
#define GL_SAMPLER_2D                     0x8B5E
#define GL_SAMPLER_3D                     0x8B5F
#define GL_SAMPLER_CUBE                   0x8B60
#define GL_SAMPLER_1D_SHADOW              0x8B61
#define GL_SAMPLER_2D_SHADOW              0x8B62
#define GL_DELETE_STATUS                  0x8B80
#define GL_COMPILE_STATUS                 0x8B81
#define GL_LINK_STATUS                    0x8B82
#define GL_VALIDATE_STATUS                0x8B83
#define GL_INFO_LOG_LENGTH                0x8B84
#define GL_ATTACHED_SHADERS               0x8B85
#define GL_ACTIVE_UNIFORMS                0x8B86
#define GL_ACTIVE_UNIFORM_MAX_LENGTH      0x8B87
#define GL_SHADER_SOURCE_LENGTH           0x8B88
#define GL_ACTIVE_ATTRIBUTES              0x8B89
#define GL_ACTIVE_ATTRIBUTE_MAX_LENGTH    0x8B8A
#define GL_FRAGMENT_SHADER_DERIVATIVE_HINT 0x8B8B
#define GL_SHADING_LANGUAGE_VERSION       0x8B8C
#define GL_CURRENT_PROGRAM                0x8B8D
#define GL_POINT_SPRITE_COORD_ORIGIN      0x8CA0
#define GL_LOWER_LEFT                     0x8CA1
#define GL_UPPER_LEFT                     0x8CA2
#define GL_STENCIL_BACK_REF               0x8CA3
#define GL_STENCIL_BACK_VALUE_MASK        0x8CA4
#define GL_STENCIL_BACK_WRITEMASK         0x8CA5

// gl 2.1 defines
#define GL_PIXEL_PACK_BUFFER              0x88EB
#define GL_PIXEL_UNPACK_BUFFER            0x88EC
#define GL_PIXEL_PACK_BUFFER_BINDING      0x88ED
#define GL_PIXEL_UNPACK_BUFFER_BINDING    0x88EF
#define GL_FLOAT_MAT2x3                   0x8B65
#define GL_FLOAT_MAT2x4                   0x8B66
#define GL_FLOAT_MAT3x2                   0x8B67
#define GL_FLOAT_MAT3x4                   0x8B68
#define GL_FLOAT_MAT4x2                   0x8B69
#define GL_FLOAT_MAT4x3                   0x8B6A
#define GL_SRGB                           0x8C40
#define GL_SRGB8                          0x8C41
#define GL_SRGB_ALPHA                     0x8C42
#define GL_SRGB8_ALPHA8                   0x8C43
#define GL_COMPRESSED_SRGB                0x8C48
#define GL_COMPRESSED_SRGB_ALPHA          0x8C49

// GL_ARB_point_parameters doesn;t define i or iv versions (http://www.opengl.org/registry/specs/ARB/point_parameters.txt)
// so for max compatibility we use that
extern PFNGLPOINTPARAMETERFARBPROC qglPointParameterf;
extern PFNGLPOINTPARAMETERFVARBPROC qglPointParameterfv;

extern PFNGLDRAWRANGEELEMENTSEXTPROC qglDrawRangeElements;
extern PFNGLLOCKARRAYSEXTPROC qglLockArrays;
extern PFNGLUNLOCKARRAYSEXTPROC qglUnlockArrays;

extern qboolean gl_arb_texture_compression;
extern qboolean gl_arb_pixel_buffer_object;
extern qboolean gl_arb_vertex_buffer_object;
extern qboolean gl_arb_point_parameters;
extern qboolean gl_ext_compiled_vertex_array;

void R_AddEntityToAlphaList (entity_t *ent);

unsigned short *R_TransferIndexes (unsigned short *src, unsigned short *dst, int numindexes, int offset);
void GL_DrawIndexedPrimitive (GLenum mode, int numIndexes, int numVertexes);
void GL_DrawPrimitive (GLenum mode, int firstvert, int numverts);
void GL_SetIndices (int indexBuffer, void *indexes);
void GL_SetStreamSource (int vbonum, int streamnum, int size, GLenum type, int stride, void *ptr);

void GL_UnbindBuffers (void);

// set these up so that we can do bitwise on them some time
#define GLSTREAM_POSITION		1
#define GLSTREAM_COLOR			2
#define GLSTREAM_TEXCOORD0		4
#define GLSTREAM_TEXCOORD1		8
#define GLSTREAM_TEXCOORD2		16

void R_ShowTrisBegin (void);
void R_ShowTrisEnd (void);
extern cvar_t r_showtris;

// this order matches a d3d fvf order which may therefore be assumed to be more optimal for hardware
typedef struct r_defaultquad_s
{
	float xyz[3];

	union
	{
		byte color[4];
		unsigned int rgba;
	};

	float st[2];
} r_defaultquad_t;

// for vertex buffer quads
typedef struct r_defaultquad2_s
{
	float xyz[3];

	union
	{
		byte color[4];
		unsigned int rgba;
	};
} r_defaultquad2_t;

extern r_defaultquad_t *r_default_quads;
extern r_defaultquad2_t *r_default_quads2;

extern int r_max_quads;
extern int r_num_quads;

extern unsigned short *r_quad_indexes;

extern qboolean gl_arb_shader_objects;

// GL 2.0 core functions
extern PFNGLBLENDEQUATIONSEPARATEPROC qglBlendEquationSeparate;
extern PFNGLDRAWBUFFERSPROC qglDrawBuffers;
extern PFNGLSTENCILOPSEPARATEPROC qglStencilOpSeparate;
extern PFNGLSTENCILFUNCSEPARATEPROC qglStencilFuncSeparate;
extern PFNGLSTENCILMASKSEPARATEPROC qglStencilMaskSeparate;
extern PFNGLATTACHSHADERPROC qglAttachShader;
extern PFNGLBINDATTRIBLOCATIONPROC qglBindAttribLocation;
extern PFNGLCOMPILESHADERPROC qglCompileShader;
extern PFNGLCREATEPROGRAMPROC qglCreateProgram;
extern PFNGLCREATESHADERPROC qglCreateShader;
extern PFNGLDELETEPROGRAMPROC qglDeleteProgram;
extern PFNGLDELETESHADERPROC qglDeleteShader;
extern PFNGLDETACHSHADERPROC qglDetachShader;
extern PFNGLDISABLEVERTEXATTRIBARRAYPROC qglDisableVertexAttribArray;
extern PFNGLENABLEVERTEXATTRIBARRAYPROC qglEnableVertexAttribArray;
extern PFNGLGETACTIVEATTRIBPROC qglGetActiveAttrib;
extern PFNGLGETACTIVEUNIFORMPROC qglGetActiveUniform;
extern PFNGLGETATTACHEDSHADERSPROC qglGetAttachedShaders;
extern PFNGLGETATTRIBLOCATIONPROC qglGetAttribLocation;
extern PFNGLGETPROGRAMIVPROC qglGetProgramiv;
extern PFNGLGETPROGRAMINFOLOGPROC qglGetProgramInfoLog;
extern PFNGLGETSHADERIVPROC qglGetShaderiv;
extern PFNGLGETSHADERINFOLOGPROC qglGetShaderInfoLog;
extern PFNGLGETSHADERSOURCEPROC qglGetShaderSource;
extern PFNGLGETUNIFORMLOCATIONPROC qglGetUniformLocation;
extern PFNGLGETUNIFORMFVPROC qglGetUniformfv;
extern PFNGLGETUNIFORMIVPROC qglGetUniformiv;
extern PFNGLGETVERTEXATTRIBDVPROC qglGetVertexAttribdv;
extern PFNGLGETVERTEXATTRIBFVPROC qglGetVertexAttribfv;
extern PFNGLGETVERTEXATTRIBIVPROC qglGetVertexAttribiv;
extern PFNGLGETVERTEXATTRIBPOINTERVPROC qglGetVertexAttribPointerv;
extern PFNGLISPROGRAMPROC qglIsProgram;
extern PFNGLISSHADERPROC qglIsShader;
extern PFNGLLINKPROGRAMPROC qglLinkProgram;
extern PFNGLSHADERSOURCEPROC qglShaderSource;
extern PFNGLUSEPROGRAMPROC qglUseProgram;
extern PFNGLUNIFORM1FPROC qglUniform1f;
extern PFNGLUNIFORM2FPROC qglUniform2f;
extern PFNGLUNIFORM3FPROC qglUniform3f;
extern PFNGLUNIFORM4FPROC qglUniform4f;
extern PFNGLUNIFORM1IPROC qglUniform1i;
extern PFNGLUNIFORM2IPROC qglUniform2i;
extern PFNGLUNIFORM3IPROC qglUniform3i;
extern PFNGLUNIFORM4IPROC qglUniform4i;
extern PFNGLUNIFORM1FVPROC qglUniform1fv;
extern PFNGLUNIFORM2FVPROC qglUniform2fv;
extern PFNGLUNIFORM3FVPROC qglUniform3fv;
extern PFNGLUNIFORM4FVPROC qglUniform4fv;
extern PFNGLUNIFORM1IVPROC qglUniform1iv;
extern PFNGLUNIFORM2IVPROC qglUniform2iv;
extern PFNGLUNIFORM3IVPROC qglUniform3iv;
extern PFNGLUNIFORM4IVPROC qglUniform4iv;
extern PFNGLUNIFORMMATRIX2FVPROC qglUniformMatrix2fv;
extern PFNGLUNIFORMMATRIX3FVPROC qglUniformMatrix3fv;
extern PFNGLUNIFORMMATRIX4FVPROC qglUniformMatrix4fv;
extern PFNGLVALIDATEPROGRAMPROC qglValidateProgram;
extern PFNGLVERTEXATTRIB1DPROC qglVertexAttrib1d;
extern PFNGLVERTEXATTRIB1DVPROC qglVertexAttrib1dv;
extern PFNGLVERTEXATTRIB1FPROC qglVertexAttrib1f;
extern PFNGLVERTEXATTRIB1FVPROC qglVertexAttrib1fv;
extern PFNGLVERTEXATTRIB1SPROC qglVertexAttrib1s;
extern PFNGLVERTEXATTRIB1SVPROC qglVertexAttrib1sv;
extern PFNGLVERTEXATTRIB2DPROC qglVertexAttrib2d;
extern PFNGLVERTEXATTRIB2DVPROC qglVertexAttrib2dv;
extern PFNGLVERTEXATTRIB2FPROC qglVertexAttrib2f;
extern PFNGLVERTEXATTRIB2FVPROC qglVertexAttrib2fv;
extern PFNGLVERTEXATTRIB2SPROC qglVertexAttrib2s;
extern PFNGLVERTEXATTRIB2SVPROC qglVertexAttrib2sv;
extern PFNGLVERTEXATTRIB3DPROC qglVertexAttrib3d;
extern PFNGLVERTEXATTRIB3DVPROC qglVertexAttrib3dv;
extern PFNGLVERTEXATTRIB3FPROC qglVertexAttrib3f;
extern PFNGLVERTEXATTRIB3FVPROC qglVertexAttrib3fv;
extern PFNGLVERTEXATTRIB3SPROC qglVertexAttrib3s;
extern PFNGLVERTEXATTRIB3SVPROC qglVertexAttrib3sv;
extern PFNGLVERTEXATTRIB4NBVPROC qglVertexAttrib4Nbv;
extern PFNGLVERTEXATTRIB4NIVPROC qglVertexAttrib4Niv;
extern PFNGLVERTEXATTRIB4NSVPROC qglVertexAttrib4Nsv;
extern PFNGLVERTEXATTRIB4NUBPROC qglVertexAttrib4Nub;
extern PFNGLVERTEXATTRIB4NUBVPROC qglVertexAttrib4Nubv;
extern PFNGLVERTEXATTRIB4NUIVPROC qglVertexAttrib4Nuiv;
extern PFNGLVERTEXATTRIB4NUSVPROC qglVertexAttrib4Nusv;
extern PFNGLVERTEXATTRIB4BVPROC qglVertexAttrib4bv;
extern PFNGLVERTEXATTRIB4DPROC qglVertexAttrib4d;
extern PFNGLVERTEXATTRIB4DVPROC qglVertexAttrib4dv;
extern PFNGLVERTEXATTRIB4FPROC qglVertexAttrib4f;
extern PFNGLVERTEXATTRIB4FVPROC qglVertexAttrib4fv;
extern PFNGLVERTEXATTRIB4IVPROC qglVertexAttrib4iv;
extern PFNGLVERTEXATTRIB4SPROC qglVertexAttrib4s;
extern PFNGLVERTEXATTRIB4SVPROC qglVertexAttrib4sv;
extern PFNGLVERTEXATTRIB4UBVPROC qglVertexAttrib4ubv;
extern PFNGLVERTEXATTRIB4UIVPROC qglVertexAttrib4uiv;
extern PFNGLVERTEXATTRIB4USVPROC qglVertexAttrib4usv;
extern PFNGLVERTEXATTRIBPOINTERPROC qglVertexAttribPointer;

extern GLuint hLiquidProgram;
extern GLuint hSkyProgram;

#define GL_LINES_ADJACENCY_ARB 0xA
#define GL_LINE_STRIP_ADJACENCY_ARB 0xB
#define GL_TRIANGLES_ADJACENCY_ARB 0xC
#define GL_TRIANGLE_STRIP_ADJACENCY_ARB 0xD
#define GL_PROGRAM_POINT_SIZE_ARB 0x8642
#define GL_MAX_VARYING_COMPONENTS_ARB 0x8B4B
#define GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS_ARB 0x8C29
#define GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER_ARB 0x8CD4
#define GL_FRAMEBUFFER_ATTACHMENT_LAYERED_ARB 0x8DA7
#define GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_ARB 0x8DA8
#define GL_FRAMEBUFFER_INCOMPLETE_LAYER_COUNT_ARB 0x8DA9

#define GL_GEOMETRY_SHADER_ARB 0x8DD9
#define GL_GEOMETRY_VERTICES_OUT_ARB 0x8DDA
#define GL_GEOMETRY_INPUT_TYPE_ARB 0x8DDB
#define GL_GEOMETRY_OUTPUT_TYPE_ARB 0x8DDC
#define GL_MAX_GEOMETRY_VARYING_COMPONENTS_ARB 0x8DDD
#define GL_MAX_VERTEX_VARYING_COMPONENTS_ARB 0x8DDE
#define GL_MAX_GEOMETRY_UNIFORM_COMPONENTS_ARB 0x8DDF
#define GL_MAX_GEOMETRY_OUTPUT_VERTICES_ARB 0x8DE0
#define GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS_ARB 0x8DE1

typedef void (APIENTRYP PFNGLFRAMEBUFFERTEXTUREARBPROC) (GLenum target, GLenum attachment, GLuint texture, GLint level);
typedef void (APIENTRYP PFNGLFRAMEBUFFERTEXTUREFACEARBPROC) (GLenum target, GLenum attachment, GLuint texture, GLint level, GLenum face);
typedef void (APIENTRYP PFNGLFRAMEBUFFERTEXTURELAYERARBPROC) (GLenum target, GLenum attachment, GLuint texture, GLint level, GLint layer);
typedef void (APIENTRYP PFNGLPROGRAMPARAMETERIARBPROC) (GLuint program, GLenum pname, GLint value);

extern PFNGLPROGRAMPARAMETERIARBPROC qglProgramParameteri;
extern PFNGLFRAMEBUFFERTEXTUREARBPROC qglFramebufferTexture;
extern PFNGLFRAMEBUFFERTEXTURELAYERARBPROC qglFramebufferTextureLayer;
extern PFNGLFRAMEBUFFERTEXTUREFACEARBPROC qglFramebufferTextureFace;

#endif	/* __GLQUAKE_H */

